package tlass.project;

import java.util.ArrayList;

import tlass.project.gui.panels.Restriction;
import tlass.project.model.Assistant;
import tlass.project.model.Instructor;
import tlass.project.model.Model;
import tlass.project.model.Position;
import tlass.project.model.Response;
import tlass.project.model.Schedule;
import tlass.project.model.Section;
import tlass.project.model.Skill;
import tlass.project.model.TimeBlock;


public class ModelHandler {
	private Model model;
	
	public ModelHandler(Model m) {
		model= m;
	}
	
	
	/**
	 * Returns a new schedule by utilizing ModelEvent
	 * @return
	 */
	public Schedule pickNewSchedule() {
		Schedule newSchedule = Scheduler.pick(model);
		if (newSchedule != null) {
			model.addSchedule(newSchedule);
			notifySchedulesListeners(new ModelEvent(model.getSchedules(), 
					newSchedule, 
					ModelEvent.ADDED));
		}
		return newSchedule;
	}

	/**
	 * Adds a new Assistant to the model by utilizing ModelEvent
	 * @return
	 */
	public Assistant addNewAssistant() {
		Assistant a = new Assistant();
		model.addAssistant(a);
		notifyAssistantsListeners(new ModelEvent(model.getAssistants(), 
				a, 
				ModelEvent.ADDED));
		return a;
	}

	/**
	 * Adds a new Section to the model
	 * @return
	 */
	public Section addNewSection() {
		Section s = new Section();
		model.addSection(s);
		notifySectionsListeners(new ModelEvent(model.getSections(), 
				s, 
				ModelEvent.ADDED));
		return s;
	}
	

	/**
	 * Adds a new TA position to the provided section.
	 * @param s
	 */
	public Position addTAPosition(Section s) {

		Position p= s.addTAPosition();
		for (Schedule aSchedule : model.getSchedules()) {
			aSchedule.addAssignment(p, null);
		}
		notifyPositionsListeners(new ModelEvent(model.getPositions(),p, ModelEvent.ADDED));
		return p;
	}
	
	/**
	 * Adds a new LA Position to the provided section
	 * @param s
	 */
	public Position addLAPosition(Section s) {
		Position p= s.addLAPosition();
		for (Schedule aSchedule : model.getSchedules()) {
			aSchedule.addAssignment(p, null);
		}
		notifyPositionsListeners(new ModelEvent(model.getPositions(),p, ModelEvent.ADDED));
		return p;
	}

	/**
	 * Adds a new Instructor to the model by utilizing ModelEvent
	 * @return
	 */
	public Instructor addNewInstructor() {
		Instructor i= new Instructor();
		model.addInstructor(i);
		notifyInstructorsListeners(new ModelEvent(model.getInstructors(), 
				i, 
				ModelEvent.ADDED));
		return i;
	}

	/**
	 * Returns the model
	 * @return
	 */
	public Object getModel() {
		return model;
	}

	/**
	 * Sets the current model to the one that was provided.
	 * @param newModel
	 */
	public void setModel(Model newModel) {
		model= newModel;
	}

	/**
	 * Adds the specified skill to the model.
	 * @param newSkill
	 */
	public void addSkill(Skill newSkill) {
		model.addSkill(newSkill);
		notifySkillsListeners(new ModelEvent(model.getSkills(), 
				newSkill, 
				ModelEvent.ADDED));
	}

	/**
	 * Adds the specified time block to the model.
	 * @param newTime
	 */
	public void addTime(TimeBlock newTime){
		model.addTime(newTime);
	}
	

	/**
	 * Removes the supplied object from the model using ModelEvent
	 * @param userObject
	 */
	public void removeResource(Object userObject) {
		if (userObject instanceof Assistant) {
			model.removeAssistant((Assistant)userObject);
			notifyAssistantsListeners(new ModelEvent(model.getAssistants(), 
													userObject, 
													ModelEvent.REMOVED));
		} else if (userObject instanceof Section) {
			model.removeSection((Section)userObject);
			notifySectionsListeners(new ModelEvent(model.getSections(), 
					userObject, 
					ModelEvent.REMOVED));
		} else if (userObject instanceof Schedule) {
			model.removeSchedule((Schedule)userObject);
			notifySchedulesListeners(new ModelEvent(model.getSchedules(), 
					userObject, 
					ModelEvent.REMOVED));
		} else if (userObject instanceof Instructor) {
			model.removeInstructor((Instructor)userObject);
			notifyInstructorsListeners(new ModelEvent(model.getInstructors(), 
					userObject, 
					ModelEvent.REMOVED));
		}
	}
	
	/**
	 * Renames the supplied object to the supplied name 
	 * @param userObject
	 * @param name
	 */
	public void renameResource(Object userObject, String name){
		if(userObject instanceof Section){
			model.renameSection((Section)userObject, name);
		}
		else if(userObject instanceof Instructor){
			model.renameInstructor((Instructor)userObject, name);
		}
		else if(userObject instanceof Assistant){
			model.renameAssistant((Assistant)userObject, name);
		}
		else if (userObject instanceof Schedule){
			model.renameSchedule((Schedule)userObject, name);
		}
		}
	
	/**
	 * Restores the schedule to the matching stored schedule, reverting any changes.
	 * @param schedule
	 * @param scheduleBackup
	 */
	public void restoreSchedule(Schedule schedule, Schedule scheduleBackup) {
		scheduleBackup.clean(model.getPositions(), model.getAssistants());
		schedule.restore(scheduleBackup);
		notifySchedulesListeners(new ModelEvent(model.getSchedules(), schedule, ModelEvent.RENAMED));
	}
	
	/**
	 * Restores the supplied assistant to the to the matching stored assistant, reverting any changes.
	 * @param assistant
	 * @param assistantBackup
	 */
	public void restoreAssistant(Assistant assistant, Assistant assistantBackup) {
		assistantBackup.clean();
		assistant.restore(assistantBackup);
		notifyAssistantsListeners(new ModelEvent(model.getAssistants(), assistant, ModelEvent.RENAMED));
	}

	/**
	 * Notifies AssistantListeners of any changes
	 * @param me
	 */
	private void notifyAssistantsListeners(ModelEvent me) {
		for (ModelListener ml : assistantsListener) {
			ml.onEvent(me);
		}
	}
	
	/**
	 * Notifies SectionListeners of any changes 
	 * @param me
	 */
	private void notifySectionsListeners(ModelEvent me) {
		for (ModelListener ml : sectionsListener) {
			ml.onEvent(me);
		}
	}
	
	/**
	 * Notifies the PositionsListeners of any changes
	 * @param me
	 */
	private void notifyPositionsListeners(ModelEvent me) {
		for (ModelListener ml : positionsListener) {
			ml.onEvent(me);
		}
	}
	
	/**
	 * Notifies the ScheduleListeners of any changes
	 * @param me
	 */
	private void notifySchedulesListeners(ModelEvent me) {
		for (ModelListener ml : schedulesListener) {
			ml.onEvent(me);
		}
	}
	
	/**
	 * Notifies the InstructorsListeners of any changes
	 * @param me
	 */
	private void notifyInstructorsListeners(ModelEvent me) {
		for (ModelListener ml : instructorsListener) {
			ml.onEvent(me);
		}
	}
	
	/**
	 * Notifies the SkillListeners of any changes
	 * @param me
	 */
	private void notifySkillsListeners(ModelEvent me) {
		for (ModelListener ml : skillsListener) {
			ml.onEvent(me);
		}
	}
	
	// ArrayLists of ModelListeners
	ArrayList<ModelListener> assistantsListener= new ArrayList<ModelListener>();
	ArrayList<ModelListener> sectionsListener= new ArrayList<ModelListener>();
	ArrayList<ModelListener> positionsListener= new ArrayList<ModelListener>();
	ArrayList<ModelListener> schedulesListener= new ArrayList<ModelListener>();
	ArrayList<ModelListener> instructorsListener= new ArrayList<ModelListener>();
	ArrayList<ModelListener> skillsListener= new ArrayList<ModelListener>();
	
	/**
	 * Put the supplied ModelListener in to the assistantsListener ArrayList
	 * @param ml
	 */
	public void addAssistantsListener(ModelListener ml) {
		assistantsListener.add(ml);
	}
	
	/**
	 * Put the supplied ModelListener in to the sectionsListener ArrayList
	 * @param ml
	 */
	public void addSectionsListener(ModelListener ml) {
		sectionsListener.add(ml);
	}
	
	/**
	 * Put the supplied ModelListener in to the positionsListener ArrayList
	 * @param ml
	 */
	public void addPositionsListener(ModelListener ml) {
		positionsListener.add(ml);
	}
	
	/**
	 * Put the supplied ModelListener in to the schedulesListener ArrayList
	 * @param ml
	 */
	public void addSchedulesListener(ModelListener ml) {
		schedulesListener.add(ml);
	}
	
	/**
	 * Put the supplied ModelListener in to the instructorsListener ArrayList
	 * @param ml
	 */
	public void addInstructorsListener(ModelListener ml) {
		instructorsListener.add(ml);
	}
	
	/**
	 * Put the supplied ModelListener in to the skillsListener ArrayList
	 * @param ml
	 */
	public void addSkillsListener(ModelListener ml) {
		skillsListener.add(ml);
	}
	
	/**
	 * Removes ALL of the ModelListeners from all of the ArrayLists
	 */
	public void removeAllListeners() {
		assistantsListener.clear();
		sectionsListener.clear();
		positionsListener.clear();
		schedulesListener.clear();
		instructorsListener.clear();
		skillsListener.clear();
	}


	/**
	 * Loads the supplied model
	 * @param newModel
	 */
	public void loadModel(Model newModel) {
		model.loadModel(newModel);
	}


	/**
	 * Sets the semester as the current semester that will be interacted with
	 * @param selection
	 */
	public void setSemester(String selection) {
		model.setSemester(selection);
	}

	/**
	 * Publishes a schedule, ensuring that all other schedules are not
	 * published.
	 * @param schedule
	 */
	public void publishSchedule(Schedule schedule) {
		for (Schedule aSchedule : model.getSchedules()) {
			aSchedule.unpublish();
			aSchedule.unfinalize();
		}
		schedule.publish();
	}
	
	/**
	 * Sets the flag that tells the system that the schedule will be published to false
	 * @param schedule
	 */
	public void unpublishSchedule(Schedule schedule) {
		schedule.unpublish();
	}
	
	/**
	 * Finalizes a schedule, ensuring that all other schedules are not
	 * finalized.
	 * @param schedule
	 */
	public void finalizeSchedule(Schedule schedule) {
		for (Schedule aSchedule : model.getSchedules()) {
			aSchedule.unpublish();
			aSchedule.unfinalize();
		}
		schedule.finalize();
	}


	/**
	 * Unpublishes the schedule if it cannot be published
	 * @param schedule
	 */
	public void unpublishIfBad(Schedule schedule) {
		if (!schedule.canPublish()) {
			schedule.unpublish();
		}
	}


	/**
	 * Changes a response to a position in a schedule. Returns whether the response was
	 * different from what was there before.
	 * @param schedule
	 * @param position
	 * @param response
	 * @return
	 */
	public boolean setResponse(Schedule schedule, Position position,
			Response response) {
		return schedule.setResponse(position, response);
	}


	/**
	 * Sets the flag in the system that says whether a Schedule is finalized to false.
	 * @param schedule
	 */
	public void unfinalizeSchedule(Schedule schedule) {
		schedule.unfinalize();
	}


	/**
	 * Adds a time that an assistant will not be able to work to the supplied assistant.
	 * @param assistant
	 * @param time
	 */
	public void addUnavailableTime(Assistant assistant, TimeBlock time) {
		assistant.addUnavailableTime(time);
	}


	/**
	 * Gets the positions that will potentially cause the Schedule from being completed.
	 * @return
	 */
	public ArrayList<Position> findProblemPositions() {
		return Scheduler.findProblemPositions(model);
	}


	/**
	 * Changes the name of the schedule to the one supplied.
	 * @param schedule
	 * @param text
	 */
	public void setScheduleName(Schedule schedule, String text) {
		schedule.setName(text);
		notifySchedulesListeners(new ModelEvent(model.getSchedules(), schedule, ModelEvent.RENAMED));
	}


	/**
	 * Changes the name of the Assistant to the one supplied.
	 * @param assistant
	 * @param text
	 */
	public void setAssistantName(Assistant assistant, String text) {
		assistant.setName(text);
		notifyAssistantsListeners(new ModelEvent(model.getAssistants(), assistant, ModelEvent.RENAMED));
	}
	
	/**
	 * Changes the name of the Instructor to the one supplied.
	 * @param instructor
	 * @param text
	 */
	public void setInstructorName(Instructor instructor, String text) {
		instructor.setName(text);
		notifyInstructorsListeners(new ModelEvent(model.getInstructors(), instructor, ModelEvent.RENAMED));
	}


	/**
	 * Sets the Assistant's username to the one upplied.
	 * @param assistant
	 * @param text
	 */
	public void setAssistantUsername(Assistant assistant, String text) {
		assistant.setUsername(text);
	}


	/**
	 * Sets the threshold of students in a section the Assistant can be assigned to 
	 * @param assistant
	 * @param value
	 */
	public void setAssistantLimit(Assistant assistant, int value) {
		assistant.setStudentsLimit(value);
	}


	/**
	 * Sets the Assistant to be either active or inactive
	 * marking whether or not they are eligable for scheduling
	 * @param assistant
	 * @param selected
	 */
	public void setAssistantActive(Assistant assistant, boolean selected) {
		assistant.setActive(selected);
	}


	/**
	 * Removes a timeblock that is unavailable from the Assistant
	 * @param assistant
	 * @param aTime
	 */
	public void removeAssistantUnavailableTime(Assistant assistant,
			TimeBlock aTime) {
		assistant.removeUnavailableTime(aTime);
	}


	/**
	 * Adds a time that an assistant will not be able to work to the supplied assistant.
	 * @param assistant
	 * @param time
	 */
	public boolean addAssistantUnavailableTime(Assistant assistant, TimeBlock time) {
		return assistant.addUnavailableTime(time);
	}


	/**
	 * Adds a skill that the Assistant possesses to that Assistant.
	 * @param assistant
	 * @param newSkillName
	 */
	public void addAssistantSkill(Assistant assistant, String newSkillName) {
		addAssistantSkill(assistant, new Skill(newSkillName));
	}

	/**
	 * Returns if added
	 * @param assistant
	 * @param newSkill
	 * @return
	 */
	public boolean addAssistantSkill(Assistant assistant, Skill newSkill) {
		if (!model.skillExists(newSkill)) {
			model.addSkill(newSkill);
			notifySkillsListeners(new ModelEvent(model.getSkills(), newSkill, ModelEvent.ADDED));
		}
		if (assistant.hasSkill(newSkill)) {
			return false;
		} else {
			assistant.addSkill(newSkill);
			return true;
		}
	}


	/**
	 * Removes a skill from an Assistant
	 * @param assistant
	 * @param selectedSkill
	 */
	public void removeAssistantSkill(Assistant assistant, Skill selectedSkill) {
		assistant.removeSkill(selectedSkill);
		if (!model.skillReferenced(selectedSkill)) {
			model.removeSkill(selectedSkill);
			notifySkillsListeners(new ModelEvent(model.getSkills(), selectedSkill, ModelEvent.REMOVED));
		}
	}
	
	/**
	 * Removes a skill that is required of an Assistant from the Section.
	 * @param section
	 * @param selectedSkill
	 */
	public void removeSectionSkill(Section section, Skill selectedSkill) {
		section.removeSkill(selectedSkill);
		if (!model.skillReferenced(selectedSkill)) {
			model.removeSkill(selectedSkill);
			notifySkillsListeners(new ModelEvent(model.getSkills(), selectedSkill, ModelEvent.REMOVED));
		}
	}

	/**
	 * Add a skill that is required of an Assistant to the Section
	 * @param section
	 * @param newSkillName
	 */
	public void addSectionSkill(Section section, String newSkillName) {
		addSectionSkill(section, new Skill(newSkillName));
	}

	/**
	 * Add a skill that is required of an Assistant to the Section
	 * also checks to see if the skill is already present before adding it
	 * @param section
	 * @param newSkill
	 * @return
	 */
	public boolean addSectionSkill(Section section, Skill newSkill) {
		if (!model.skillExists(newSkill)) {
			model.addSkill(newSkill);
			notifySkillsListeners(new ModelEvent(model.getSkills(), newSkill, ModelEvent.ADDED));
		}
		if (section.hasSkill(newSkill)) {
			return false;
		} else {
			section.addSkill(newSkill);
			return true;
		}
	}


	/**
	 * Restores an Instructor's data to the data stored in the backup.
	 * @param instructor
	 * @param instructorBackup
	 */
	public void restoreInstructor(Instructor instructor, Instructor instructorBackup) {
		instructorBackup.clean(model.getAssistants());
		instructor.restore(instructorBackup);
		notifyInstructorsListeners(new ModelEvent(model.getInstructors(), instructor, ModelEvent.RENAMED));
	}



	public void setInstructorUsername(Instructor instructor, String text) {
		instructor.setId(text);
	}



	/**
	 * Restores a section from a backup
	 * @param section
	 * @param sectionBackup
	 */
	public void restoreSection(Section section, Section sectionBackup) {
		sectionBackup.clean(model.getAssistants(), model.getSkills(), model.getInstructors());
		section.restore(sectionBackup);
		notifySectionsListeners(new ModelEvent(model.getSections(), section, ModelEvent.RENAMED));
	}


/**
 * Removes a specific TA position from the model, updating references
 * @param position
 */
	public void removeTAPosition(Position position) {
		model.removeTAPosition(position);
		for (Schedule aSchedule : model.getSchedules()) {
			aSchedule.unassignPosition(position);
		}
		notifyPositionsListeners(new ModelEvent(model.getPositions(), position, ModelEvent.REMOVED));
	}

	/**
	 * Removes a specific LA position from the model, updating references
	 * @param position
	 */
	public void removeLAPosition(Position position) {
		model.removeLAPosition(position);
		for (Schedule aSchedule : model.getSchedules()) {
			aSchedule.unassignPosition(position);
		}
		notifyPositionsListeners(new ModelEvent(model.getPositions(), position, ModelEvent.REMOVED));
	}


/**
 * Sets a sections id
 * @param section
 * @param text
 */
	public void setSectionId(Section section, String text) {
		section.setId(text);
		notifySectionsListeners(new ModelEvent(model.getSections(), section, ModelEvent.RENAMED));
	}


/**
 * Sets a sections location
 * @param section
 * @param text
 */
	public void setSectionLocation(Section section, String text) {
		section.setLocation(text);
	}


/**
 * Sets a sections instructor
 * @param section
 * @param selectedItem
 */
	public void setSectionInstructor(Section section, Instructor selectedItem) {
		section.setInstructor(selectedItem);
	}


/**
 * Set a sections enrolled maximum
 * @param section
 * @param text
 */
	public void setSectionEnrolledMax(Section section, String text) {
		try {
			section.setEnrolledMax(Integer.parseInt(text));
		} catch (Exception e){
			// do nothign;
		}
	}

/**
 * Remove any timeblocks with a certain day from a section
 * @param section
 * @param day
 */
	public void removeSectionTimeDay(Section section, int day) {
		section.removeDay(day);
	}



	public void addSectionTime(Section section, int day, int start, int end) {
		section.addDayForce(day, start, end);
	}



	public void setSectionTime(Section section, int start, int end) {
		section.setTime(start, end);
	}



	public void setPositionRestriction(Position position,
			Restriction restriction, Object data) {
		Section s= position.getSection();
		if (restriction == Restriction.MULTIPLE) {
			s.setForbiddenAssistants(position, (ArrayList<Assistant>)data);
			s.setRequiredAssistants(position, null);
		} else if (restriction == Restriction.SINGLE) {
			s.setForbiddenAssistants(position, new ArrayList<Assistant>());
			s.setRequiredAssistants(position, (Assistant)data);
		} else {
			s.setForbiddenAssistants(position, new ArrayList<Assistant>());
			s.setRequiredAssistants(position, null);
		}
	}
}
