package tlass.project;

import java.util.ArrayList;

import tlass.project.model.Assistant;
import tlass.project.model.Constraint;
import tlass.project.model.Model;
import tlass.project.model.Position;
import tlass.project.model.Schedule;

import JaCoP.core.IntVar;
import JaCoP.core.Store;
import JaCoP.search.DepthFirstSearch;
import JaCoP.search.IndomainMin;
import JaCoP.search.InputOrderSelect;
import JaCoP.search.Search;
import JaCoP.search.SelectChoicePoint;

public class Scheduler {
	
	/**
	 * Returns an ArrayList of positions cannot be assigned.
	 * @param model
	 * @return
	 */
	public static ArrayList<Position> findProblemPositions(Model model) {
		// Create a variable for each slot
		Position[] allPositions = model.getPositions();
		Assistant[] assistants = model.getAssistantsArray();
		
		if (allPositions.length == 0) return null;
		if (assistants.length == 0) return null;

		ArrayList<Position> problemPositions = new ArrayList<Position>();
		
		for (Position aPosition : allPositions) {
			// Setup Constraint Network
			Store store = new Store();
			
			int values= assistants.length;
			Position[] positions = {aPosition};
			IntVar[] variables = {new IntVar(store, aPosition.getId(), 0, values-1)}; 
	
			// Apply all constraints
			for (Constraint aConstraint : model.getConstraints()) {
				if (aConstraint.isEnabled()) {
					aConstraint.imposeConstraint(store, variables, positions, assistants);
				}
			}
			
			// Solve the constraint network
			Search<IntVar> search= new DepthFirstSearch<IntVar>();
			search.setPrintInfo(false);
			SelectChoicePoint<IntVar> select = 
	            new InputOrderSelect<IntVar>(store, variables, 
	                                         new IndomainMin<IntVar>());
			boolean result = search.labeling(store, select);
			if (!result) {
				problemPositions.add(aPosition);
			}
		}
		return problemPositions;
	}

	/**
	 * Creates a schedule by taking all of the positions and assistants
	 * and applying the constraints and returns a list of pairings
	 * that abides by the constraints.
	 * @param model
	 * @return
	 */
	public static Schedule pick(Model model){
		// Setup Constraint Network
		Store store = new Store();
		
		// Create a variable for each slot
		Position[] positions = model.getPositions();
		Assistant[] assistants = model.getAssistantsArray();
		
		if (positions.length == 0) return null;
		if (assistants.length == 0) return null;
		
		
		IntVar[] variables = new IntVar[positions.length];
		int values= assistants.length;
		for (int i= 0; i < variables.length; i+= 1) {
			variables[i]= new IntVar(store, positions[i].getId(), 0, values-1);
		}

		// Apply all constraints
		for (Constraint aConstraint : model.getConstraints()) {
			if (aConstraint.isEnabled()) {
				aConstraint.imposeConstraint(store, variables, positions, assistants);
			}
		}
		
		// Solve the constraint network
		Search<IntVar> search= new DepthFirstSearch<IntVar>();
		search.setPrintInfo(false);
		SelectChoicePoint<IntVar> select = 
            new InputOrderSelect<IntVar>(store, variables, 
                                         new IndomainMin<IntVar>());
		boolean result = search.labeling(store, select);
		
		if (!result) {
			return null;
		}
		
		/*for (IntVar aVar : variables) {
			System.out.println(aVar.id() +" : " + model.getAssistant(aVar.value()));
		}*/
		
		// Iterate over the variables to get the solved assignments
		Schedule schedule= new Schedule();

		for (int i=0; i < variables.length; i+= 1) {
			Position position= positions[i];
			int assistantIndex= variables[i].value();
			Assistant assistant= model.getAssistant(assistantIndex);
			schedule.addAssignment(position, assistant);
		}

		// Immediately report the results for debug purposes
		/*for (IntVar aVar : variables) {
			System.out.println(aVar.id() +" : " + model.getAssistant(aVar.value()));
		}*/
		
		return schedule;
	}
	
		
	
}
