package tlass.project;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;


import java.sql.Blob;

import tlass.project.exceptions.DbConnectionException;
import tlass.project.model.Model;

public class SemesterHandler {

	/**
	 * Saves all of the data from the semester to the DataStore
	 * @param name
	 * @param model
	 * @throws DbConnectionException
	 * @throws SQLException
	 * @throws IOException
	 */
	public void saveSemester(String name, Model model) throws DbConnectionException, SQLException, IOException {
		Connection connection = new DataStoreConnection().get();
		String command= "UPDATE sessions SET session = ? " +
				"WHERE semester = ?";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setBlob(1, blobify(model));
		statement.setString(2, name);
		statement.executeUpdate();
		statement.close();
		connection.close();
	}
	
	/**
	 * Sets the selected semester as the active semester.
	 * @param name
	 * @param active
	 * @throws DbConnectionException
	 * @throws SQLException
	 * @throws IOException
	 */
	public void setActiveSemester(String name, boolean active) throws DbConnectionException, SQLException, IOException {
		int activeVal;
		if (active) activeVal= 1;
		else 		activeVal= 0;
		Connection connection = new DataStoreConnection().get();
		String command= "UPDATE sessions " +
				"SET active = IF(semester = ?, ?, 0) ";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setString(1, name);
		statement.setInt(2, activeVal);
		statement.executeUpdate();
		statement.close();
		connection.close();
	}

	/**
	 * Loads the semesters from the DataStore
	 * @param semester
	 * @return
	 * @throws DbConnectionException
	 * @throws SQLException
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public Model loadSemester(String semester) throws DbConnectionException, SQLException, IOException, ClassNotFoundException {
		Connection connection = new DataStoreConnection().get();
		String command= "SELECT session, active FROM sessions WHERE semester = ?";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setString(1, semester);
		ResultSet results= statement.executeQuery();
		Model model= null;
		if (results.next()) {
			Blob sessionBlob= results.getBlob("session");
			if (sessionBlob != null) {
				model= (Model) deblobify(sessionBlob);
			}
			int active= results.getInt("active");
			if (active == 0) {
				model.setActive(false);
			} else {
				model.setActive(true);
			}
		}
		results.close();
		statement.close();
		connection.close();
		return model;
	}

	/**
	 * Retrieves the names of all of the semesters from the DataStore.
	 * @return
	 * @throws SQLException
	 * @throws DbConnectionException
	 */
	public ArrayList<String> getSemesters() throws SQLException, DbConnectionException {
		Connection connection = new DataStoreConnection().get();
		String command= "SELECT semester FROM sessions";
		PreparedStatement statement= connection.prepareStatement(command);
		ResultSet results= statement.executeQuery();
		ArrayList<String> semesters= new ArrayList<String>();
		while (results.next()) {
			semesters.add(results.getString("semester"));
		}
		results.close();
		statement.close();
		connection.close();
		return semesters;
	}

	/**
	 * Delete the supplied semester from the dataStore
	 * @param name
	 * @throws DbConnectionException
	 * @throws SQLException
	 */
	public void deleteSemester(String name) throws DbConnectionException, SQLException {
		Connection connection = new DataStoreConnection().get();
		String command= "DELETE FROM sessions WHERE semester = ?";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setString(1, name);
		statement.executeUpdate();
		statement.close();
		connection.close();
	}

	/**
	 * Adds a semester of the supplied name to the supplied model.
	 * @param name
	 * @param model
	 * @throws DbConnectionException
	 * @throws SQLException
	 * @throws IOException
	 */
	public void addSemester(String name, Model model) throws DbConnectionException, SQLException, IOException {
		Connection connection = new DataStoreConnection().get();
		String command= "INSERT INTO sessions (semester, session, active)" +
				"Values (?, ?, 0)";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setString(1, name);
		statement.setBlob(2, blobify(model));
		statement.executeUpdate();
		statement.close();
		connection.close();
	}

	/**
	 * Consumes a blob and converts it into an object if it can
	 * @param sessionBlob
	 * @return
	 * @throws SQLException
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private Object deblobify(Blob sessionBlob) throws SQLException, IOException, ClassNotFoundException {
		byte[] sessionBytes= sessionBlob.getBytes(1, (int)sessionBlob.length());
		return DataStoreConnection.readBytes(sessionBytes);
	}

	/**
	 * Consumes an object and turns it into a blob representation
	 * @param anObject
	 * @return
	 * @throws IOException
	 */
	private InputStream blobify(Object anObject) throws IOException {
		byte[] sessionBytes= DataStoreConnection.writeBytes(anObject);
		InputStream sessionStream= new ByteArrayInputStream(sessionBytes);
		return sessionStream;
	}

	/**
	 * Takes the selected semester and clones it
	 * @param name
	 * @param oldSemester
	 * @throws DbConnectionException
	 * @throws SQLException
	 */
	public void copySemester(String name, String oldSemester) throws DbConnectionException, SQLException {
		Connection connection = new DataStoreConnection().get();
		String command= "INSERT INTO sessions " +
				"(SELECT ?, session, 0 FROM sessions WHERE semester = ?)";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setString(1, name);
		statement.setString(2, oldSemester);
		statement.executeUpdate();
		statement.close();
		connection.close();
	}

}
