package tlass.project;


import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Time;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map.Entry;

import tlass.project.exceptions.DbConnectionException;
import tlass.project.model.Assistant;
import tlass.project.model.Model;
import tlass.project.model.Position;
import tlass.project.model.Response;
import tlass.project.model.Schedule;
import tlass.project.model.TimeBlock;


public class WebformHandler {
	
	private Controller controller;
	
	/**
	 * create a webformhandler
	 * @param Controller
	 */
	public WebformHandler(Controller c) {
		controller= c;
	}
	
	/**
	 * Retrieves the Assignments from the System for the specified semester
	 * @param semester the semester to get assignments for
	 * @return ArrayList<Pair<String, String>> the assignments
	 * @throws DbConnectionException
	 * @throws SQLException
	 */
	public ArrayList<Pair<String, String>> getAssignments(String semester) throws DbConnectionException, SQLException {
		Connection connection = new DataStoreConnection().get();
		String command= "SELECT username, id, ta, place FROM publishedassignments " +
				"WHERE semester = ?";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setString(1, semester);
		ResultSet results= statement.executeQuery();
		ArrayList<Pair<String, String>> assignments= new ArrayList<Pair<String, String>>();
		while (results.next()) {
			Model model = controller.getModel();
			String position = Position.buildId(results.getString("id"),results.getInt("ta"),results.getInt("place"));
			String username = results.getString("username");
			Pair<String, String> assignment = new Pair<String, String>(position, username); 
			assignments.add(assignment);
		}
		results.close();
		statement.close();
		connection.close();
		return assignments;
	}
	
	/**
	 * Notifies the assistant of their assignment 
	 * @param assistant
	 */
	public void notifyAssistant(Assistant assistant) {
		//TODO
	}
	
	/**
	 * Notifies multiple assistants of their assignments
	 * @param usersChanged
	 */
	private void notifyAssistants(String[] usersChanged) {
		// TODO Auto-generated method stub
		
	}
	
	/**
	 * Removes published assignments from the webform.
	 * @param String semester to remove published assignments from
	 * @throws DbConnectionException
	 * @throws SQLException
	 */
	private void removePublishedAssignments(String semester) throws DbConnectionException, SQLException {
		Connection connection = new DataStoreConnection().get();
		String command= "DELETE FROM publishedassignments " +
				"WHERE semester = ?";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setString(1, semester);
		statement.executeUpdate();
		statement.close();
		connection.close();
	}
	
	/**
	 * Adds published assignments of the specified semester to the webform
	 * @param String semester to add published assignments to
	 * @param Schedule to add published assignments to
	 * @throws DbConnectionException
	 * @throws SQLException
	 */
	private void addPublishedAssignments(String semester, Schedule schedule) throws DbConnectionException, SQLException {
		Connection connection = new DataStoreConnection().get();
		String command= "INSERT INTO publishedassignments " +
				"VALUES (?, ? , ?, ?, ?, ?, ?, ?, ?)";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setNull(1, java.sql.Types.INTEGER);
		statement.setString(2, semester);
		HashMap<Position, Assistant> assignments = schedule.getAssignments();
		for (Entry<Position, Assistant> pair : assignments.entrySet()) {
			Position position= pair.getKey();
			Assistant assistant = pair.getValue();
			statement.setString(3, assistant.getUsername());
			statement.setString(4, position.getSectionId());
			statement.setInt(5, position.getType());
			statement.setInt(6, position.getOrder());
			statement.setString(7, position.getDescription());
			statement.setInt(8, 0);
			statement.setString(9, "");
			statement.executeUpdate();
		}
		statement.close();
		connection.close();
	}

	/**
	 * Removes the List of finalized assignments from webform.
	 * @param String semester to remove finalized assignments from
	 * @throws DbConnectionException
	 * @throws SQLException
	 */
	private void removeFinalizedAssignments(String semester) throws DbConnectionException, SQLException {
		Connection connection = new DataStoreConnection().get();
		String command= "DELETE FROM finalizedassignments " +
				"WHERE semester = ?";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setString(1, semester);
		statement.executeUpdate();
		statement.close();
		connection.close();
	}
	
	/**
	 * Adds the finalized assignments to the webform.
	 * @param String semester to add finalized assignments to
	 * @param Schedule to add finalized assignments to
	 * @throws DbConnectionException
	 * @throws SQLException
	 */
	private void addFinalizedAssignments(String semester, Schedule schedule) throws DbConnectionException, SQLException {
		Connection connection = new DataStoreConnection().get();
		String command= "INSERT INTO finalizedassignments " +
				"VALUES (?, ? , ?, ?, ?, ?, ?)";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setNull(1, java.sql.Types.INTEGER);
		statement.setString(2, semester);
		HashMap<Position, Assistant> assignments = schedule.getAssignments();
		for (Entry<Position, Assistant> pair : assignments.entrySet()) {
			Position position= pair.getKey();
			Assistant assistant = pair.getValue();
			statement.setString(3, assistant.getUsername());
			statement.setString(4, position.getSectionId());
			statement.setInt(5, position.getType());
			statement.setInt(6, position.getOrder());
			statement.setString(7, position.getDescription());
			statement.executeUpdate();
		}
		statement.close();
		connection.close();
	}
	
	/**
	 * Removes the published schedule from the webForm
	 * @param String semester to unpublish
	 * @throws DbConnectionException
	 * @throws SQLException
	 */
	public void unpublishSchedule(String semester) throws DbConnectionException, SQLException{
		removePublishedAssignments(semester);
	}
	
	/**
	 * Publishes the schedule to the webform.
	 * @param String semester to publish a schedule for
	 * @param Schedule that will be published
	 * @return String[] of all the users that were changed
	 * @throws DbConnectionException
	 * @throws SQLException
	 */
	public String[] publishSchedule(String semester, Schedule schedule) throws DbConnectionException, SQLException{
		ArrayList<Pair<String, String>> assignments= getAssignments(semester);
		String[] usersChanged;
		if (assignments.isEmpty()) {
			usersChanged= schedule.getAssistants();
		} else {
			usersChanged= schedule.compare(assignments);
		}
		notifyAssistants(usersChanged);
		removePublishedAssignments(semester);
		addPublishedAssignments(semester, schedule);
		return usersChanged;
	}

	/**
	 * Notifies the Assistants of their assignments and finalizes the assignments
	 * @param String semester that a schedule is being finalized for
	 * @param Schedule that is being finalzed
	 * @return String[] of all the users changed
	 * @throws DbConnectionException
	 * @throws SQLException
	 */
	public String[] finalizeSchedule(String semester, Schedule schedule) throws DbConnectionException, SQLException{
		String[] usersChanged= schedule.getAssistants();
		notifyAssistants(usersChanged);
		removePublishedAssignments(semester);
		removeFinalizedAssignments(semester);
		addFinalizedAssignments(semester, schedule);
		return usersChanged;
	}
	
	/**
	 * Retieves the responses from the webform and updates the published schedule accordingly.
	 * @param String semester to get new responses for
	 * @param Schedule to get new responses for
	 * @return ArrayList<Pair<Position, Response>> of the new responses
	 * @throws SQLException
	 * @throws DbConnectionException
	 */
	public ArrayList<Pair<Position, Response>> getNewResponses(String semester, Schedule schedule) throws SQLException, DbConnectionException{
		Connection connection = new DataStoreConnection().get();
		String command= "SELECT id, ta, place, username, accept, comment " +
				"FROM publishedassignments " +
				"WHERE semester = ?";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setString(1, semester);
		ResultSet results= statement.executeQuery();
		ArrayList<Pair<Position, Response>> responses= new ArrayList<Pair<Position,Response>>();
		while (results.next()) {
			Model model = controller.getModel();
			String positionId = Position.buildId(results.getString("id"),results.getInt("ta"),1+results.getInt("place"));
			String username = results.getString("username");
			Position position = model.getPositionById(positionId);
			Assistant assistant = model.getAssistantByUsername(username);
			if (position != null && assistant != null && schedule.isAssigned(position, assistant)) {
				int accept = results.getInt("accept");
				String comment = results.getString("comment");
				Response response = new Response();
				switch (accept) {
				case 0: response.unseen(); break;
				case 1: response.reject(comment); break;
				case 2: default: response.accept(); break;
				}
				Pair<Position, Response> pair = new Pair<Position, Response>(position, response);
				responses.add(pair);
			}
		}
		results.close();
		statement.close();
		connection.close();
		return responses;
	}
	
	/**
	 * Gets all of the unavailable times for assistants for the supplied semester.
	 * @param String semester to get unavailable assistants for
	 * @return ArrayList<String> of all the unavailable assistants
	 * @throws SQLException
	 * @throws DbConnectionException
	 */
	public ArrayList<String> getUnavailableAssistants(String semester) throws SQLException, DbConnectionException {
		Connection connection = new DataStoreConnection().get();
		String command= "SELECT username " +
				"FROM assistantsunavailables " +
				"WHERE semester = ?";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setString(1, semester);
		ResultSet results= statement.executeQuery();
		ArrayList<String> users= new ArrayList<String>();
		while (results.next()) {
			String username= results.getString("username");
			users.add(username);
		}
		results.close();
		statement.close();
		connection.close();
		return users;
	}
	
	/**
	 * Gets the unavalable times for a specific assistant during the supplied semester.
	 * @param String semester to get unavailable times for
	 * @param String user to get unavailable times for
	 * @return ArrayList<TimeBlock> times that are unavailable
	 * @throws SQLException
	 * @throws DbConnectionException
	 */
	public ArrayList<TimeBlock> getUnavailableTimes(String semester, String user) throws SQLException, DbConnectionException {
		Connection connection = new DataStoreConnection().get();
		String command= "SELECT username, startDay, startTime, endDay, endTime, comment " +
				"FROM assistantsunavailables " +
				"WHERE semester = ? " +
				"AND username = ?";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setString(1, semester);
		statement.setString(2, user);
		ResultSet results= statement.executeQuery();
		ArrayList<TimeBlock> times= new ArrayList<TimeBlock>();
		while (results.next()) {
			Time startTime = results.getTime("startTime");
			int startDay = results.getInt("startDay");
			Time endTime = results.getTime("endTime");
			int endDay= results.getInt("endDay");
			String comment= results.getString("comment");
			TimeBlock time = new TimeBlock(startTime, startDay, endTime, endDay, comment);
			times.add(time);
		}
		results.close();
		statement.close();
		connection.close();
		return times;
	}
	
	/**
	 * Gets all of the unavailable times for all of the assistants during the supplied semester.
	 * @param String semester to get all the unavailable times for
	 * @return ArrayList<Pair<String, TimeBlock>> all the unavailable times
	 * @throws SQLException
	 * @throws DbConnectionException
	 */
	public ArrayList<Pair<String, TimeBlock>> getAllUnavailableTimes(String semester) throws SQLException, DbConnectionException {
		Connection connection = new DataStoreConnection().get();
		String command= "SELECT username, startDay, startTime, endDay, endTime, comment " +
				"FROM assistantsunavailables " +
				"WHERE semester = ? " +
				"ORDER BY username ASC";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setString(1, semester);
		ResultSet results= statement.executeQuery();
		ArrayList<Pair<String, TimeBlock>> times= new ArrayList<Pair<String, TimeBlock>>();
		while (results.next()) {
			Time startTime = results.getTime("startTime");
			int startDay = results.getInt("startDay");
			Time endTime = results.getTime("endTime");
			int endDay= results.getInt("endDay");
			String comment= results.getString("comment");
			String username= results.getString("username");
			TimeBlock time = new TimeBlock(startTime, startDay, endTime, endDay, comment);
			times.add(new Pair(username, time));
		}
		results.close();
		statement.close();
		connection.close();
		return times;
	}

	/**
	 * Removes an unavailable time
	 * @param String semester to remove unavailable times from
	 * @throws DbConnectionException
	 * @throws SQLException
	 */
	public void removeUnavailableTimes(String semester) throws DbConnectionException, SQLException {
		Connection connection = new DataStoreConnection().get();
		String command= "DELETE FROM assistantsunavailables " +
				"WHERE semester = ?";
		PreparedStatement statement= connection.prepareStatement(command);
		statement.setString(1, semester);
		statement.executeUpdate();
		statement.close();
		connection.close();
	}
}
