/* SchedulePanel.java
 * CISC475 Project
 * Group 02
 */

package tlass.project.gui.panels;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.sql.SQLException;
import java.util.ArrayList;

import javax.swing.BoxLayout;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import tlass.project.*;
import tlass.project.exceptions.DbConnectionException;
import tlass.project.gui.MainWindow;
import tlass.project.model.Assistant;
import tlass.project.model.Position;
import tlass.project.model.Response;
import tlass.project.model.Schedule;
import tlass.project.model.Section;


public class SchedulePanel extends ResourcePanel {
	private static final long serialVersionUID = 1L;
	private Schedule schedule;
	private Schedule scheduleBackup;
	private WebformHandler webformHandler;
	private MainWindow mainWindow;
	
	private JToolBar toolBar;
	private PublishButton publishButton;
	private FinalizeButton finalizeButton;
	private ScheduleTable scheduleTable;

	public final static int COLUMN_SECTION = 0;
	public final static int COLUMN_POSITION = 1;
	public final static int COLUMN_ASSISTANT = 2;
	public final static int COLUMN_RESPONSE = 3;
	private String[] columnNames = { "Section", "Position", "Assistant",
			"Response" };
	private JTextField scheduleNameBox;

	/**
	 * Constructor of a Schedule Panel.
	 * @param mw
	 * @param c
	 * @param s
	 */
	public SchedulePanel(MainWindow mw, Controller c, Schedule s) {
		super(c, new BorderLayout(), s);
		this.schedule = s;
		this.scheduleBackup = s.copy();
		this.mainWindow = mw;
		this.webformHandler = c.getWebformHandler();
		this.setLayout(new BorderLayout());

		modelHandler.addAssistantsListener(new ModelListener() {
			@Override
			public void onEvent(ModelEvent event) {
				scheduleTable.clearTable();
				scheduleTable.buildTable();
				scheduleTable.revalidate();
				scheduleTable.repaint();
				if (event.getType() == ModelEvent.REMOVED) {
					modelHandler.unpublishIfBad(schedule);
				}
				publishButton.refresh();
				finalizeButton.refresh();
			}
		});

		modelHandler.addPositionsListener(new ModelListener() {
			@Override
			public void onEvent(ModelEvent event) {
				scheduleTable.clearTable();
				scheduleTable.buildTable();
				scheduleTable.revalidate();
				scheduleTable.repaint();
				if (event.getType() != ModelEvent.RENAMED) {
					modelHandler.unpublishIfBad(schedule);
				}
				publishButton.refresh();
				finalizeButton.refresh();
			}
		});

		modelHandler.addSectionsListener(new ModelListener() {
			@Override
			public void onEvent(ModelEvent event) {
				scheduleTable.clearTable();
				scheduleTable.buildTable();
				scheduleTable.revalidate();
				scheduleTable.repaint();
				if (event.getType() == ModelEvent.REMOVED) {
					modelHandler.unpublishIfBad(schedule);
				}
				publishButton.refresh();
				finalizeButton.refresh();
			}
		});

		// Create the options toolbar
		this.toolBar = new JToolBar();
		
		JButton saveButton= new JButton("Save and Close");
		toolBar.add(saveButton);
		saveButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {			
				mainWindow.closeTab(schedule);
			}
		});

		JButton cancelButton = new JButton("Cancel Changes");
		toolBar.add(cancelButton);
		cancelButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				modelHandler.restoreSchedule(schedule, scheduleBackup);
				mainWindow.closeTab(schedule);
			}
		});

		// Publish button becomes the "Audit" button when the current schedule
		// is the published one
		publishButton = new PublishButton();
		toolBar.add(publishButton);

		// Finalize Button
		finalizeButton = new FinalizeButton();
		toolBar.add(finalizeButton);
		
		scheduleNameBox = new JTextField(10);
		scheduleNameBox.setText(schedule.getName());
		scheduleNameBox.getDocument().addDocumentListener(new DocumentListener() {
			@Override
			public void changedUpdate(DocumentEvent arg0) {}

			@Override
			public void insertUpdate(DocumentEvent arg0) {
				update();
			}

			@Override
			public void removeUpdate(DocumentEvent arg0) {
				update();
			}
			
			public void update() {
				modelHandler.setScheduleName(schedule, scheduleNameBox.getText());
			}
		});
		toolBar.add(scheduleNameBox);

		this.add(toolBar, BorderLayout.NORTH);
		toolBar.setFloatable(false);

		this.scheduleTable = new ScheduleTable();
		JScrollPane tableScroller = new JScrollPane(scheduleTable,
				JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
				JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);

		this.add(tableScroller, BorderLayout.CENTER);
	}
	/**
	 * audits the model as the user requested
	 * @throws SQLException
	 * @throws DbConnectionException
	 */
	protected void auditSchedule() throws SQLException, DbConnectionException {
		ArrayList<Pair<Position,Response>> responses= webformHandler.getNewResponses(model.getSemester(), schedule);
		boolean changed= false;
		for (Pair<Position,Response> pair : responses) {
			Position position= pair.getFirst();
			Response response= pair.getSecond();
			boolean change= modelHandler.setResponse(schedule, position, response);
			changed |= change;
		}
		
		if (changed) {
			scheduleTable.refreshResponseButtons();
		} else {
			String message= "No assistants have responded to their assignment recently.";
			JOptionPane.showMessageDialog(null, message, "No new responses", JOptionPane.INFORMATION_MESSAGE);
		}
	}
	/**
	 * finalize this schedule
	 * @throws DbConnectionException
	 * @throws SQLException
	 */
	protected void finalizeSchedule() throws DbConnectionException, SQLException {
		String[] affectedUsers = webformHandler.finalizeSchedule(model.getSemester(), schedule);
		modelHandler.finalizeSchedule(schedule);
		scheduleTable.refreshResponseButtons();
		String[] message = {"Your schedule was finalized. The following email addresses were notified.",
							Auxiliary.join(affectedUsers, ", ")};
		JOptionPane.showMessageDialog(null, message, "Schedule Finalized", JOptionPane.INFORMATION_MESSAGE);
	}

	/**
	 * publish this schedule
	 * @throws DbConnectionException
	 * @throws SQLException
	 */
	protected void publishSchedule() throws DbConnectionException, SQLException {
		String[] affectedUsers = webformHandler.publishSchedule(model.getSemester(), schedule);
		modelHandler.publishSchedule(schedule);
		scheduleTable.refreshResponseButtons();
		String[] message = {"Your schedule was published. The following email addresses were notified.",
							Auxiliary.join(affectedUsers, ", ")};
		JOptionPane.showMessageDialog(null, message, "Schedule Published", JOptionPane.INFORMATION_MESSAGE);
	}
	
	private class FinalizeButton extends JButton {
		/**
		 * check to make sure that the user wants to proceed without all assignments being approved or overridden
		 * @return
		 */
		private boolean confirmFinalizationOverride() {
			String[] messages= {"Not all the assignments have been approved or overridden.",
					"Finalized anyway?"};
			int result = JOptionPane
					.showConfirmDialog(null,
							messages,
							"Confirm Finalization", JOptionPane.YES_NO_OPTION,
							JOptionPane.QUESTION_MESSAGE);
			return result == 0;
		}
		
		/**
		 * create a finalize button
		 */
		public FinalizeButton() {
			refresh();
			setText("Finalize");
			addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent arg0) {
					if (!schedule.isApproved()) {
						boolean reallySure= confirmFinalizationOverride();
						if (!reallySure) return;
					}
					try {
						finalizeSchedule();
						refresh();
					} catch (Exception e) {
						e.printStackTrace();
						Auxiliary.showConnectionError();
					}
				}
			});
		}
		
		/**
		 * refresh the schedule
		 */
		public void refresh() {
			if (schedule.isFinalized()) {
				setEnabled(false);
			} else {
				setEnabled(schedule.canPublish());
			}
		}
	}
	
	private class PublishButton extends JButton {
		/**
		 * create a publish button
		 */
		public PublishButton() {
			refresh();
			addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent arg0) {
					try {
						if (schedule.isPublished()) {
							auditSchedule();
						} else {
							publishSchedule();
							refresh();
						}
					} catch (Exception e) {
						e.printStackTrace();
						Auxiliary.showConnectionError();
					}
				}
			});
		}
		/**
		 * refresh the schedule
		 */
		public void refresh() {
			if (schedule.isPublished()) {
				setText("Audit");
				setEnabled(true);
			} else {
				setText("Publish");
				setEnabled(schedule.canPublish());
			}
		}
	}

	public class ScheduleTable extends JPanel {
		BoxLayout layout;
		
		private ArrayList<ResponseButton> responseButtons;

		/**
		 * create a schedule table
		 */
		public ScheduleTable() {
			layout = new BoxLayout(this, BoxLayout.Y_AXIS);
			this.setLayout(layout);
			this.setBackground(Color.WHITE);
			responseButtons= new ArrayList<ResponseButton>();
			buildTable();
		}

		/**
		 * refresh the schedule
		 */
		public void refreshResponseButtons() {
			for (ResponseButton aButton : responseButtons) {
				aButton.refresh();
			}
		}

		/**
		 * clear the table
		 */
		public void clearTable() {
			responseButtons.clear();
			this.removeAll();
		}

		/**
		 * build the table
		 */
		public void buildTable() {
			ArrayList<Section> sortedSectionList = model.getSortedSections();
			for (Section aSection : sortedSectionList) {
				// Section Label
				JLabel sectionLabel = new JLabel(aSection.toString());
				sectionLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
				add(sectionLabel);

				ArrayList<Position> sortedPositionList = aSection
						.getSortedPositions();
				for (Position aPosition : sortedPositionList) {
					Response aResponse = schedule.getResponse(aPosition);
					JPanel aRow = new JPanel();
					aRow.setBackground(Color.WHITE);

					// Position Label
					JLabel positionLabel = new JLabel(
							aPosition.getPositionName());
					aRow.add(positionLabel);

					// Assistant Combo Box
					DefaultComboBoxModel assistantModel = new DefaultComboBoxModel();
					for (Assistant a : model.getAssistants()) {
						assistantModel.addElement(a);
					}
					JComboBox assistantComboBox = new JComboBox(assistantModel);
					assistantComboBox.setSelectedItem(schedule
							.getAssignment(aPosition));
					AssistantComboBoxListener acbl = new AssistantComboBoxListener(
							aPosition, aResponse);
					assistantComboBox.addActionListener(acbl);
					aRow.add(assistantComboBox);

					// Response Button
					ResponseButton responseButton = new ResponseButton(
							schedule.getResponse(aPosition));
					acbl.setResponseButton(responseButton);
					aRow.add(responseButton);
					responseButtons.add(responseButton);

					// Add entire row
					add(aRow);
				}
			}
		}
		
	

		/**
		 * This class handles events from the assistantComboBox
		 * 
		 * Can you believe it? Anonymous classes in java c)an't have
		 * constructors! So I either had to make a bunch of complicated instance
		 * variables, or make this a named, nested class. What a piece of
		 * work...
		 * 
		 * @author acbart
		 * 
		 */
		private class AssistantComboBoxListener implements ActionListener {
			Position position;
			Response response;
			private ResponseButton responseButton;

			/**
			 * create an assistant combo box listener
			 * @param Position of the assistant
			 * @param Response from the model
			 */
			public AssistantComboBoxListener(Position p, Response r) {
				position = p;
				response = r;
			}

			/**
			 * set the response button as the given button
			 * @param ResponseButton
			 */
			public void setResponseButton(ResponseButton responseButton) {
				this.responseButton = responseButton;
			}

			/**
			 * Fired when the AssistantBox is modified; changes the assigned assistant.
			 */
			@Override
			public void actionPerformed(ActionEvent event) {
				JComboBox assistantBox = (JComboBox) event.getSource();
				Assistant newAssistant = (Assistant) (assistantBox)
						.getSelectedItem();
				if (schedule.getAssignment(position) != newAssistant) {
					schedule.addAssignment(position, newAssistant);
					modelHandler.unpublishSchedule(schedule);
					modelHandler.unfinalizeSchedule(schedule);
					publishButton.refresh();
					finalizeButton.refresh();
					response.unseen();
					responseButton.refresh();
				}
			}
		}

		private class ResponseButtonListener implements ActionListener {
			private Response response;

			/**
			 * create the listener for the response button
			 * @param Response from the model
			 */
			public ResponseButtonListener(Response r) {
				response = r;
			}

			@Override
			public void actionPerformed(ActionEvent event) {
				ResponseButton rb = (ResponseButton) event.getSource();
				switch (response.getValue()) {
				case Response.UNSEEN:
					askOverride(rb);
					break;
				case Response.REJECTED:
					askRejectedOverride(rb);
					break;
				case Response.OVERRIDDEN:
					askUnseen(rb);
					break;
				}
			}

			/**
			 * ask the user if they want to override unseen schedules
			 * @param ResponseButton 
			 */
			private void askUnseen(ResponseButton rb) {
				int result = JOptionPane
						.showConfirmDialog(null,
								"Do you want to undo this override?",
								"Undo Override", JOptionPane.YES_NO_OPTION,
								JOptionPane.QUESTION_MESSAGE);
				if (result == 0) {
					response.unseen();
					rb.refresh();
				}
			}

			/**
			 * ask the user if they want to override the rejection
			 * @param ResponseButton
			 */
			private void askRejectedOverride(ResponseButton rb) {
				String[] messages = {
						"The Assistant has rejected this assignment for the reason below.",
						response.getComment(),
						"Do you want override this rejection?" };
				int result = JOptionPane.showConfirmDialog(null, messages,
						"Override Rejection", JOptionPane.YES_NO_OPTION,
						JOptionPane.QUESTION_MESSAGE);
				if (result == 0) {
					response.override();
					rb.refresh();
				}
			}

			/**
			 * ask the user if they want to override the acceptance
			 * @param ResponseButton
			 */
			private void askOverride(ResponseButton rb) {
				int result = JOptionPane.showConfirmDialog(null,
						"Do you want to override this acceptance?",
						"Force Acceptance", JOptionPane.YES_NO_OPTION,
						JOptionPane.QUESTION_MESSAGE);
				if (result == 0) {
					response.override();
					rb.refresh();
				}
			}
		}

		private class ResponseButton extends JButton {
			private Response response;
			private Color regularColor;

			/**
			 * create the response button
			 * @param Response
			 */
			public ResponseButton(Response response) {
				super(response.toString());
				regularColor = getBackground();
				this.response = response;
				refresh();
				this.addActionListener(new ResponseButtonListener(response));
			}

			/**
			 * refresh the button
			 */
			public void refresh() {
				setText(response.toString());
				setEnabled(!response.isAccepted() && !response.isFinalized());
				if (response.isRejected()) {
					setBackground(Color.PINK);
				} else if (response.isUnseen()) {
					setBackground(Color.ORANGE);
				} else {
					setBackground(regularColor);
				}
			}
		}
	}

	/**
	 * Get the schedule
	 * @return Schedule
	 */
	public Schedule getSchedule() {
		return schedule;
	}

	/**
	 * Set the schedule
	 * @param Schedule
	 */
	public void setSchedule(Schedule s) {
		schedule = s;
	}
}
