/* Schedule.java
 * CISC475 Project
 * Group 02
 */

package tlass.project.model;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import tlass.project.Pair;




public class Schedule implements Serializable {
	
	//variables for new system design
	private HashMap<Position, Assistant> assignments;
	private HashMap<Position, Response> responses;
	private boolean published;
	private boolean finalized;
	private String name;
	
	public Schedule() {
		assignments= new HashMap<Position, Assistant>();
		responses= new HashMap<Position, Response>();
		name= "Untitled Schedule";
	}
	
	/**
	 * Returns a copy of the Schedule, including it's internal data
	 * structures (assignments and responses). Note that although the
	 * structures are different (different reference), the references they
	 * *contain* are the same.
	 * @return
	 */
	public Schedule copy() {
		Schedule clone = new Schedule();
		clone.setPublished(published);
		clone.setFinalized(finalized);
		clone.setName(name);
		clone.setAssignments((HashMap<Position, Assistant>)assignments.clone());
		clone.setResponses((HashMap<Position, Response>)responses.clone());
		return clone;
	}
	
	/**
	 * Consumes the positions and assistants in the system and checks if this
	 * schedule references resources that are missing. If a reference is not found,
	 * then it is removed.
	 * @param positions
	 * @param assistants
	 */
	public void clean(ArrayList<Position> positions, ArrayList<Assistant> assistants) {
		for (Position anOp : assignments.keySet()) {
			if (positions.contains(anOp)) {
				Assistant a= assignments.get(anOp);
				if (a != null && !assistants.contains(a)) {
					unassign(anOp);
				}
			} else {
				unassignPosition(anOp);
			}
		}
	}
	
	/**
	 * Removes all of the positions and assistants in the supplied array of positions and ArrayList of asssistants
	 * @param positions
	 * @param assistants
	 */
	public void clean(Position[] positions, ArrayList<Assistant> assistants) {
		ArrayList<Position> positionsList = new ArrayList<Position>(Arrays.asList(positions));
		clean(positionsList, assistants);
	}

	/**
	 * Overwrites a schedule based on a given backup. This allows us to keep
	 * the reference the same (so anyone with a reference to this Schedule won't
	 * get confused) while modifying the internal data. 
	 * @param backup
	 */
	public void restore(Schedule backup) {
		published= backup.isPublished();
		finalized= backup.isFinalized();
		name= backup.getName();
		assignments= backup.getAssignments();
		responses= backup.getResponses();
	}

	/**
	 * Adds a new assignment (with accompanying "unseen" response).
	 * @param position
	 * @param assistant
	 */
	public void addAssignment(Position position, Assistant assistant) {
		if (assignments.containsKey(position)) {
			assignments.put(position, assistant);
			resetResponse(position);
		} else {
			assignments.put(position, assistant);
			responses.put(position, new Response());
		}
	}

	/**
	 * Has this schedule been finalized. A finalized schedule is one
	 * that is fully approved and sent out to users. Only one schedule
	 * can be finalized at a time.
	 * @return
	 */
	public boolean isFinalized() {
		return finalized;
	}

	/**
	 * Has this schedule been published. A published schedule is one
	 * that has been sent to users for review. Only one schedule can
	 * be published at a time.
	 * @return
	 */
	public boolean isPublished() {
		return published;
	}

	/**
	 * Checks whether a position has been filled.
	 * @param position
	 * @return
	 */
	public boolean filled(Position position) {
		if (assignments.containsKey(position)) {
			return (assignments.get(position) != null);
		} else {
			return false;
		}
	}

	/** 
	 * Returns the assistant assigned to a given position
	 * @param position
	 * @return
	 */
	public Assistant getAssignment(Position position) {
		return assignments.get(position);
	}

	/**
	 * Returns the response given to this schedule by the assistant
	 * assigned to that position.
	 * @param position
	 * @return
	 */
	public Response getResponse(Position position) {
		return responses.get(position);
	}

	/**
	 * Simple accessor for assignments.
	 * @return
	 */
	public HashMap<Position, Assistant> getAssignments() {
		return assignments;
	}

	/**
	 * Removes any references to the assistant from the schedule.
	 * Additionally, resets the response for any positions that
	 * are affected. 
	 * @param a Assistant to be removed
	 */
	public void unassignAssistant(Assistant a) {
		for (Position anOp : assignments.keySet()) {
			if (assignments.get(anOp) == a) {
				unassign(anOp);
			}
		}
	}

	/**
	 * Removes the specified position from the schedule.
	 * @param p
	 */
	public void unassignPosition(Position p) {
		assignments.remove(p);
		responses.remove(p);
	}

	/**
	 * Removes the assignment of an Assistant to a Position. If the assignment did not
	 * exist before, it will be created, but be blank.
	 * @param p
	 */
	public void unassign(Position p) {
		assignments.put(p, null);
		if (responses.containsKey(p)) {
			responses.get(p).reset();
		} else {
			resetResponse(p);
		}
	}
	
	/**
	 * Removes any references to the section from the schedule.
	 * @param s
	 */
	public void unassignSection(Section s) {
		Iterator<Entry<Position, Assistant>> asit = assignments.entrySet().iterator();
		while (asit.hasNext()) {
			Map.Entry pairs = (Map.Entry)asit.next();
			if (((Position) pairs.getKey()).getSection() == s) {
				asit.remove();
			}
		}
	}

	/**
	 * Returns the Assistant assigned to a given Position represented
	 * as a String. The position's representation is assumed to be based
	 * on it's ID.
	 * @param position
	 * @return
	 */
	private Assistant getAssignment(String position) {
		for (Position anOp : assignments.keySet()) {
			if (anOp.getId().equals(position)) {
				return assignments.get(anOp);
			}
		}
		return null;
	}

	/**
	 * Given the String representation of a position (it's ID) and an
	 * Assistant (their username), check if this matches what is
	 * stored in the schedule. 
	 * @param position
	 * @param username
	 * @return
	 */
	private boolean checkWrongAssignment(String position, String username) {
		for (Entry<Position, Assistant> pair : assignments.entrySet()) {
			if (pair.getKey().getId().equals(position)) {
				return (!pair.getValue().getUsername().equals(username));
			}
		}
		return false;
	}

	/**
	 * Returns a string representation of the file, suitable for publishing
	 * and finalization.
	 * @return
	 */
	public String exportToString() {
		String flattened= "username,position,description\n";
		for (Entry<Position, Assistant> assignment : assignments.entrySet()) {
			Assistant assistant = assignment.getValue();
			Position position= assignment.getKey(); 
			Section section= position.getSection();
			
			String username= assistant.getUsername();
			String positionName= position.getId();
			String miscData= section.describe();
			
			flattened+= username+","+positionName+","+miscData+"\n";
		}
		return flattened.trim();
	}

	/**
	 * Simple mutator for the name
	 * @param name
	 */
	public void setName(String name) {
		this.name= name;
	}

	/**
	 * Returns the name of the schedule. For a more complete representation of
	 * a Schedule (one that could be converted back into a Schedule object),
	 * see exportToString().
	 */
	public String toString() {
		return this.name;
	}

	/**
	 * Returns an array of strings containing the usernames of all the 
	 * Assistants that have been assigned in this schedule.
	 * @return
	 */
	public String[] getAssistants() {
		HashSet<String> usernames = new HashSet<String>();
		for (Entry<Position, Assistant> assignment : assignments.entrySet()) {
			Assistant assistant= assignment.getValue(); 
			String username= assistant.getUsername();
			usernames.add(username);
		}
		return usernames.toArray(new String[usernames.size()]);
	}

	/**
	 * Consumes a schedule represented by an arraylist of Assignments, and returns
	 * the username (Strings) of all the assistants who have had their assignment 
	 * changed. An Assignment is a pair of strings representing the ID of a Position
	 * and the username of an Assistant.
	 * @param oldAssignments
	 * @return
	 */
	public String[] compare(ArrayList<Pair<String, String>> oldAssignments) {
		HashSet<String> changedUsers= new HashSet<String>();
		for (Pair<String, String> anAssignment : oldAssignments) {
			String position= anAssignment.getFirst();
			String username= anAssignment.getSecond();
			if (checkWrongAssignment(position, username)) {
				changedUsers.add(username);
				Assistant oldAssistant= getAssignment(position);
				changedUsers.add(oldAssistant.getUsername());
			}
		}
		return changedUsers.toArray(new String[changedUsers.size()]);
	}

	/**
	 * Simple accessor for the name of a Schedule.
	 * @return
	 */
	public String getName() {
		return name;
	}

	/**
	 * Sets whether a schedule has been published.
	 * @param published
	 */
	public void setPublished(boolean published) {
		this.published = published;
	}

	/**
	 * Sets whether a schedule has been finalized.
	 * @param finalized
	 */
	public void setFinalized(boolean finalized) {
		this.finalized = finalized;
	}

	/**
	 * Simple accessor for the responses of this schedule.
	 * @return
	 */
	public HashMap<Position, Response> getResponses() {
		return responses;
	}

	/**
	 * Simple mutator for the responses of this schedule.
	 * @param responses
	 */
	public void setResponses(HashMap<Position, Response> responses) {
		this.responses = responses;
	}

	/**
	 * Simple mutator for the assignments of this schedule.
	 * @param assignments
	 */
	public void setAssignments(HashMap<Position, Assistant> assignments) {
		this.assignments = assignments;
	}

	/**
	 * Publishes this schedule
	 */
	public void publish() {
		published= true;
		unseeResponses();
	}

	/**
	 * Changes all responses so that they haven't been seen.
	 */
	public void unseeResponses() {
		for (Entry<Position, Response> pairs : responses.entrySet()) {
			Response r= pairs.getValue();
			r.unseen();
		}
	}
	
	/**
	 * Changes all responses so that they have been finalized.
	 */
	public void finalizeResponses() {
		for (Entry<Position, Response> pairs : responses.entrySet()) {
			Response r= pairs.getValue();
			r.finalize();
		}
	}

	/**
	 * Unpublishes this schedule
	 */
	public void unpublish() {
		published= false;
	}

	/**
	 * Finalizes this schedule
	 */
	public void finalize() {
		finalized= true;
		finalizeResponses();
	}

	/**
	 * Unfinalizes this schedule
	 */
	public void unfinalize() {
		finalized= false;
	}

	/**
	 * A schedule can be published if all of it's positions have been filled.
	 * @return
	 */
	public boolean canPublish() {
		boolean allFilledSoFar= true;
		for (Position position : assignments.keySet()) {
			allFilledSoFar &= filled(position);
		}
		return allFilledSoFar;
	}

	/**
	 * Checks whether a given assistant is used at all in the schedule.
	 * @param assistant
	 * @return
	 */
	public boolean uses(Assistant assistant) {
		return assignments.containsValue(assistant);
	}

	/**
	 * Checks whether a given position is used at all in the schedule
	 * @param position
	 * @return
	 */
	public boolean uses(Position position) {
		return assignments.containsKey(position);
	}

	/**
	 * Checks whether a given Section is used at all in the schedule
	 * @param section
	 * @return
	 */
	public boolean uses(Section section) {
		for (Position aPosition : section.getPositions()) {
			if (uses(aPosition)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Returns whether the given assistant is assigned to that position
	 * @param position
	 * @param assistant
	 * @return
	 */
	public boolean isAssigned(Position position, Assistant assistant) {
		return (assignments.get(position) == assistant);
	}

	/**
	 * Changes the response for a certain assignment. Returns whether a change was made
	 * @param position
	 * @param response
	 * @return 
	 */
	public boolean setResponse(Position position, Response response) {
		Response r= responses.get(position);
		if (r.sameAs(response)) {
			return false;
		} else {
			r.setFrom(response);
			return true;
		}
	}
	
	/**
	 * Changes the response to unseen. Returns whether a change was made.
	 * @param position
	 * @param response
	 * @return 
	 */
	public boolean resetResponse(Position position) {
		Response r= responses.get(position);
		if (r.isUnseen()) {
			return false;
		} else {
			r.unseen();
			return true;
		}
	}
	

	/**
	 * Checks to see is a schedule can be published and is approved
	 * If so it signals that the schedule can be finalized. 
	 * @return
	 */
	public boolean canFinalize() {
		return (canPublish() && isApproved());
	}


	/**
	 * Whether a schedule has had each response to an assignment set to
	 * accept or override. 
	 * @return
	 */
	public boolean isApproved() {
		for (Entry<Position, Response> pairs : responses.entrySet()) {
			Response r= pairs.getValue();
			if (!r.isAccepted() && !r.isOverridden()) {
				return false;
			}
		}
		return true;
	}

}
