package tlass.project.model;

import java.io.Serializable;
import java.sql.Time;
import java.util.Calendar;

public class TimeBlock implements Serializable {

	public final static int SUNDAY = 0;
	public final static int MONDAY = 1;
	public final static int TUESDAY = 2;
	public final static int WEDNESDAY = 3;
	public final static int THURSDAY = 4;
	public final static int FRIDAY = 5;
	public final static int SATURDAY = 6;
	public final static int DAYS = 7;

	private final static int NOON= 720;
	// THIS IS THE DUMBEST CONSTANT EVER.
	private final static long EPOCH_TIME_OFFSET= 21596400;
	private final static long MILLISECONDS_IN_MINUTE= 60000;
	private static final int MIDNIGHT = 0;

	/**
	 * Time is measured in minutes during the day, with 12am being 0 and 11:59pm being 1439
	 * Day is 0 - 6
	 */
	private int startTime;
	private int startDay;
	private int endTime;
	private int endDay;
	private String description;

	/**
	 * Constructor
	 * @param startTime time in total minutes of the day
	 * @param startDay number day of the week
	 * @param endTime time in total minutes of the day
	 * @param endDay number day of the week
	 */
	public TimeBlock(int startTime, int startDay, int endTime, int endDay) {
		this.startTime= startTime;
		this.startDay= startDay;
		this.endTime= endTime;
		this.endDay= endDay;
	}

	/**
	 * Constructor
	 * @param day number day of the week
	 * @param sm start minute
	 * @param sh start hour
	 * @param em end minute
	 * @param eh end hour
	 */
	public TimeBlock(int day, int sm, int sh, int em, int eh) {
		this.startTime = convertTime(sm, sh);
		this.endTime= convertTime(em, eh);
		this.startDay= day;
		this.endDay= day;
	}

	/**
	 * Consumes minutes and hours (in military time format) and returns the
	 * total minutes of the day 
	 * @param minutes
	 * @param hours
	 * @return
	 */
	public static int convertTime(int minutes, int hours) {
		return 60*hours + minutes;
	}

	/**
	 * Constructor
	 * @param startTime a Java Time
	 * @param startDay number day of the week
	 * @param endTime a Java Time
	 * @param endDay number day of the week
	 */
	public TimeBlock(Time startTime, int startDay, Time endTime, int endDay) {
		this.startTime= convertTimeToTime(startTime);
		this.startDay= startDay;
		this.endTime= convertTimeToTime(endTime);
		this.endDay= endDay;
	}

	/**
	 * Constructor
	 * @param startTime a Java Time (unix epoch)
	 * @param startDay number day of the week
	 * @param endTime a Java Time (unix epoch)
	 * @param endDay number day of the week
	 * @param comment
	 */
	public TimeBlock(Time startTime, int startDay, Time endTime,
			int endDay, String comment) {
		this.startTime= convertTimeToTime(startTime);
		this.startDay= startDay;
		this.endTime= convertTimeToTime(endTime);
		this.endDay= endDay;
		this.description= comment;
	}

	/**
	 * Constructor
	 * @param day number day of the week
	 * @param start a time in minutes total of hte day
	 * @param end a time in minutes total of the day
	 */
	public TimeBlock(int day, int start, int end) {
		startTime= start;
		startDay= day;
		endTime= end;
		endDay= day;
		description= "";
	}

	/**
	 * Returns the hour that this started
	 * @return
	 */
	public int getStartHour() {
		return convertTimeToHour(startTime);
	}
	/**
	 * Returns the hour that this ended
	 * @return
	 */
	public int getEndHour() {
		return convertTimeToHour(endTime);
	}

	/**
	 * Returns the minute this started
	 * @return
	 */
	public int getStartMinute() {
		return convertTimeToMinute(startTime);
	}

	/**
	 * Returns the minute this ended
	 * @return
	 */
	public int getEndMinute() {
		return convertTimeToMinute(endTime);
	}

	/**
	 * Returns whether it started am or pm
	 * @return
	 */
	public String getStartMerdian() {
		return convertTimeToMeridian(startTime);
	}
	/**
	 * Returns whether it ended am or pm
	 * @return
	 */
	public String getEndMerdian() {
		return convertTimeToMeridian(endTime);
	}

	/**
	 * Returns the string representation of the start day
	 * @return
	 */
	public String getStartDayName() {
		return convertDayToName(startDay);
	}
	/**
	 * Returns the string representation of the end day
	 * @return
	 */
	public String getEndDayName() {
		return convertDayToName(endDay);
	}

	/**
	 * Consumes a java time and returns it in total minutes of the day
	 * @param time
	 * @return
	 */
	public static int convertTimeToTime(Time time) {
		Calendar c= Calendar.getInstance();
		c.setTime(time);
		return c.get(Calendar.HOUR_OF_DAY)*60 + c.get(Calendar.MINUTE);
	}

	/**
	 * Turns a number day into it's human readable version
	 * @param day
	 * @return
	 */
	public static String convertDayToName(int day) {
		switch (day) {
		case SUNDAY: return "Sunday";
		case MONDAY: return "Monday";
		case TUESDAY: return "Tuesday";
		case WEDNESDAY: return "Wednesday";
		case THURSDAY: return "Thursday";
		case FRIDAY: return "Friday";
		case SATURDAY: return "Saturday";
		default: return "Unknown Day ("+day+")";
		}
	}

	/**
	 * Returns the meridian of a time in minutes total
	 * @param time
	 * @return
	 */
	public static String convertTimeToMeridian(int time) {
		if (time >= NOON) return "PM";
		else 			  return "AM";
	}

	/**
	 * Returns the hour (non-military) of a time in minutes total
	 * @param time
	 * @return
	 */
	public static int convertTimeToHour(int time) {
		int hours= (time / 60) % 12;
		if (hours == 0) {
			return 12;
		} else {
			return hours;
		}
	}

	/**
	 * Returns the minutes of a time in minutes total
	 * @param time
	 * @return
	 */
	public static int convertTimeToMinute(int time) {
		return time % 60;
	}

	/**
	 * Returns it as a string
	 */
	public String toString() {
		return getTimeString();
	}

	/**
	 * Check if this time block has any intersections between it's times and days
	 * @param other
	 * @return
	 */
	public boolean intersect(TimeBlock other) {
		boolean intersectTime = !((other.getEndTime() < startTime) || (endTime < other.getStartTime()));
		boolean intersectDay = !((other.getEndDay() < endDay) || (endDay < other.getStartDay()));
		return intersectTime && intersectDay;
	}

	/**
	 * Setter for start time (uses minute-total)
	 * @param start
	 */
	public void setStartTime(int start) {
		this.startTime = start;
	}
	/**
	 * Getter for start time (uses minute-total)
	 * @return
	 */
	public int getStartTime() {
		return startTime;
	}
	/**	
	 * Setter for end time (uses minute-total)
	 * @param end
	 */
	public void setEndTime(int end) {
		this.endTime = end;
	}
	/**
	 * Getter for end time (uses minute-total
	 * @return
	 */
	public int getEndTime() {
		return endTime;
	}
	/**
	 * Setter for description
	 * @param description
	 */
	public void setDescription(String description) {
		this.description = description;
	}
	/**
	 * Getter for description
	 * @return
	 */
	public String getDescription() {
		return description;
	}

	/**
	 * Setter for startDay
	 * @param startDay
	 */
	public void setStartDay(int startDay) {
		this.startDay = startDay;
	}

	/**
	 * Getter for startDay
	 * @return
	 */
	public int getStartDay() {
		return startDay;
	}

	/**
	 * SEtter for end day
	 * @param endDay
	 */
	public void setEndDay(int endDay) {
		this.endDay = endDay;
	}

	/**
	 * Getter for end day
	 * @return
	 */
	public int getEndDay() {
		return endDay;
	}

	/**
	 * Returns a nicely formatted string representation of the start time
	 * @return
	 */
	public String getStartString() {
		return getStartHour()+":"+
		padTime(getStartMinute(), 2)+" "+
		getStartMerdian();
	}

	/**
	 * Returns a nicely formatted string representation of the end time
	 * @return
	 */
	public String getEndString() {
		return getEndHour()+":"+
		padTime(getEndMinute(),2)+" "+getEndMerdian();
	}

	/**
	 * Returns a nicely formatted string representation of the entire time
	 * @return
	 */
	public String getTimeString() {
		if (startDay == endDay) {
			return getStartDayName()+" "+getStartString() +" to "+getEndString();
		} else {
			return getStartDayName()+" "+getStartString() +" to "+getEndDayName()+" "+getEndString();
		}
	}

	/**
	 * Pads zeroes onto a number if it is less than 10
	 * @return
	 */
	public static String padTime(int t, int n) {
		return String.format("%"+n+"s", t).replace(' ', '0');
	}

	/**
	 * Check if two timeblocks are equivalent
	 * @param time
	 * @return
	 */
	public boolean sameTime(TimeBlock time) {
		return (time.getStartTime() == this.getStartTime() &&
				time.getEndTime() == this.getEndTime() &&
				time.getStartDay() == this.getStartDay() &&
				time.getEndDay() == this.getEndDay());
	}

	/**
	 * Sets the start time using minutes and hours
	 * @param sm minutes
	 * @param sh hours
	 */
	public void setStartTime(int sm, int sh) {
		startTime= convertTime(sm, sh);
	}

	/**
	 * Sets the end time using minutes and hours
	 * @param sm minutes
	 * @param sh hours
	 */
	public void setEndTime(int em, int eh) {
		endTime= convertTime(em, eh);
	}

	/**
	 * Creates a minute-total time 
	 * @param m minutes
	 * @param h hours
	 * @param startMeridian either "PM" or "AM"
	 * @return
	 */
	public static int makeTime(int m, int h, String startMeridian) {
		if (startMeridian.equals("am") || startMeridian.equals("AM")) {
			return convertTime(m, h);
		} else {
			return convertTime(m, 12+h);
		}
	}

	/**
	 * Returns the end hour as a string
	 * @return
	 */
	public String getEndHourString() {
		return Integer.toString(getEndHour());
	}
	/**
	 * Returns the start hour as a string
	 * @return
	 */
	public String getStartHourString() {
		return Integer.toString(getStartHour());
	}
	/**
	 * Returns the end minute as a string
	 * @return
	 */
	public String getEndMinuteString() {
		return padTime(getEndMinute(),2);
	}
	/**
	 * Returns the start minute as a string
	 * @return
	 */
	public String getStartMinuteString() {
		return padTime(getStartMinute(),2);
	}

}